#include <qcolor.h>
#include <qtooltip.h>
#include <qapplication.h>
#include <qdir.h>
#include <qfiledialog.h>
#if QT_VERSION >= 300
  #include <qdesktopwidget.h>
#endif

#include "controlpanel.h"
#include "configuration.h"
#include "waskin.h"
#include "gbutton.h"
#include "zipextract.h"
#include "playlistmanager.h"

ControlPanel::ControlPanel( QWidget *parent, const char *name, WFlags f )
             :QWidget(parent, name, f | Qt::WStyle_Customize | Qt::WStyle_NoBorder | Qt::WMouseNoMask | Qt::WGroupLeader),
              defaultSkin(this),
              btPrev(false, this),
              btPlay(false, this),
              btPause(false, this),
              btStop(false, this),
              btNext(false, this),
              btEject(false, this),
              btEq(true, this),
              btPl(true, this),
              btShuffle(true, this),
              btRepeat(true, this),
              txTitle(this),
              txKbps(this),
              txKHz(this),
              slVolume(this),
              slBalance(this),
              slPosition(this),
              btMin(false, this),
              btMax(true, this),
              btClose(false, this),

              timer(this),
              dirTimer(this),
              dir(Configuration::getSkinDir()),
              mainMenu(this),
              skinMenu(this),
              countMode(false),
              volumeEdit(false),
              balanceEdit(false),
              seekEdit(false),
              mouseDown(false),
              elapsed(0),
              status(),
              mpdCom(Configuration::getServerIP(), Configuration::getServerPort(), false, this),
              playlistMpdCom(Configuration::getServerIP(), Configuration::getServerPort(), true, this),
              plw(&playlistMpdCom, this),
              configDialog(),
              skin(NULL)
{
  setCaption(STR_TITLE " " LANG_VERSION " " __VERSION);

  // If the skin directory does not exist, create it.
  if (!dir.exists())
    dir.mkdir(dir.absPath());

  // Build the skin menu
  slotRebuildSkinMenu();

  // Build the main menu
  mainMenu.insertItem(LANG_SETTINGS, this, SLOT(slotShowSettings()));
  mainMenu.insertItem(LANG_SKIN, &skinMenu);
  mainMenu.insertSeparator();
  mainMenu.insertItem(LANG_ABOUT, this, SLOT(slotAbout()));

  // Apply the skin
  skin->apply();

  // Set the tooltips
  QToolTip::add(&btPrev,     LANG_PREVIOUS);
  QToolTip::add(&btPlay,     LANG_PLAY);
  QToolTip::add(&btPause,    LANG_PAUSE);
  QToolTip::add(&btStop,     LANG_STOP);
  QToolTip::add(&btNext,     LANG_NEXT);
  QToolTip::add(&btShuffle,  LANG_SHUFFLE);
  QToolTip::add(&btRepeat,   LANG_REPEAT);
  QToolTip::add(&btPl,       LANG_PLAYLIST);
  QToolTip::add(&slVolume,   LANG_VOLUME);
  QToolTip::add(&slBalance,  LANG_CROSSFADE);
  QToolTip::add(&slPosition, LANG_SEEK );

  // Connect important slots
  dbgConnect(&btPrev, SIGNAL(clicked()), &mpdCom, SLOT(previous()) );
  dbgConnect(&btPlay, SIGNAL(clicked()), &mpdCom, SLOT(play()) );
  dbgConnect(&btPause, SIGNAL(clicked()), &mpdCom, SLOT(pause()) );
  dbgConnect(&btStop, SIGNAL(clicked()), &mpdCom, SLOT(stop()) );
  dbgConnect(&btNext, SIGNAL(clicked()), &mpdCom, SLOT(next()) );
  dbgConnect(&btShuffle, SIGNAL(toggled(bool)), &mpdCom, SLOT(random(bool)) );
  dbgConnect(&btRepeat, SIGNAL(toggled(bool)), &mpdCom, SLOT(repeat(bool)) );
  dbgConnect(&btPl, SIGNAL(toggled(bool)), this, SLOT(slotShowPlaylist(bool)) );
  dbgConnect(&btEject, SIGNAL(toggled(bool)), this, SLOT(slotEjectPressed(bool)) );

  dbgConnect(&btClose, SIGNAL(clicked()), this, SLOT(close()));
  dbgConnect(&btMin, SIGNAL(clicked()), this, SLOT(showMinimized()));

  dbgConnect(&slVolume, SIGNAL(sliderPressed()), this, SLOT(slotLockVolume()));
  dbgConnect(&slVolume, SIGNAL(sliderMoved(int)), this, SLOT(slotReverseVolume(int)));
  dbgConnect(&slVolume, SIGNAL(sliderReleased()), this, SLOT(slotReleaseVolume()));

  dbgConnect(&slBalance, SIGNAL(sliderPressed()), this, SLOT(slotLockBalance()));
  dbgConnect(&slBalance, SIGNAL(sliderMoved(int)), this, SLOT(slotReverseBalance(int)));
  dbgConnect(&slBalance, SIGNAL(sliderReleased()), this, SLOT(slotReleaseBalance()));

  dbgConnect(&slPosition, SIGNAL(sliderPressed()), this, SLOT(slotLockSeek()));
  dbgConnect(&slPosition, SIGNAL(sliderMoved(int)), this, SLOT(slotSetSeek(int)));
  dbgConnect(&slPosition, SIGNAL(sliderReleased()), this, SLOT(slotReleaseSeek()));

  dbgConnect(&timer, SIGNAL(timeout()), this, SLOT(slotUpdateDisplay()));
  dbgConnect(&dirTimer, SIGNAL(timeout()), this, SLOT(slotRebuildSkinMenu()));
  dbgConnect(&mpdCom, SIGNAL(statusUpdate(const MpdCom::Status &)), this, SLOT(slotReceiveUpdate(const MpdCom::Status &)));
  dbgConnect(&mpdCom, SIGNAL(error(const QString &)), this, SLOT(slotError(const QString &)));
  timer.start(100);
  dirTimer.start(3000);

  // Show me
  show();
  mpdCom.connectCom();

  // Some defaults
  move(Configuration::getGeometry("controlpanel").topLeft());
  actualX = pos().x();
  actualY = pos().y();
  slotShowPlaylist(Configuration::getPlaylistVisible());
  countMode = Configuration::getCountMode();
  slBalance.setValue(50);
  txTitle.setText(STR_TITLE " " LANG_VERSION " " __VERSION " - " __AUTHORS);
}

void ControlPanel::setSkin(WASkin *skin)
{
  if (this->skin != skin)
    this->skin->unload();

  this->skin = skin;

  skin->setMainWidget(*this);
  skin->setControlButtons(btPrev, btPlay, btPause, btStop, btNext, btEject);
  skin->setOptionButtons(btEq, btPl, btShuffle, btRepeat);
  skin->setTextView(txTitle, txKbps, txKHz);
  skin->setSliders(slVolume, slBalance, slPosition);
  skin->setTitleButtons(btMin, btMax, btClose);

  repaint();
}

void ControlPanel::slotUpdateDisplay(void)
{
  // Omit bugs
  if ((size().width() != 275) ||
      (size().height() != 116))
  {
    resize(275, 116);
  }

  if (status.state == MpdCom::PLAY)
    elapsed++;

  if (status.state == MpdCom::STOP)
  {
    slPosition.setValue(0);
  }
  else
  {
    if (elapsed > status.songLength)
      elapsed = status.songLength;

    if (seekEdit == false)
      slPosition.setValue(elapsed);
  }

  repaint(false);
}

void ControlPanel::slotReceiveUpdate(const MpdCom::Status & status)
{
  this->status = status;
  QString text = "";

  if (status.songInfo)
  {
    if (status.songInfo->title.length() > 0)
      text += status.songInfo->title;
    if (status.songInfo->artist.length() > 0)
      text += " [" + status.songInfo->artist + "]";
    if (text.length() == 0)
    {
      QString song, artist;

      if (getSongArtist(status.songInfo->filename, song, artist))
      {
        if (song.length() > 0)
          text += song;
        if (artist.length() > 0)
          text += " [" + artist + "]";
      }
      else
        text = status.songInfo->filename;
    }
  }

  btShuffle.setOn(status.random);
  btRepeat.setOn(status.repeat);

  if (status.state != MpdCom::STOP)
  {
    setCaption(text + "(" STR_TITLE " " LANG_VERSION " " __VERSION ")");
    txKbps.setText(QString::number(status.bitrate));
    txKHz.setText(QString::number(status.samplerate / 1000));
    if ((volumeEdit == false) && (balanceEdit == false))
      txTitle.setText(text);

    if (volumeEdit == false)
    {
      QToolTip::add(&slVolume, LANG_VOLUME ": " + QString::number(status.volume) + " %");
      slVolume.setValue(status.volume);
    }

    if (balanceEdit == false)
    {
      QToolTip::add(&slBalance, LANG_CROSSFADE ": " + QString::number(status.xfade * 0.1) + " sec");
      slBalance.setValue(status.xfade);
    }

    if (seekEdit == false)
    {
      slPosition.setMax(status.songLength);
      if ((elapsed / 10) != (status.elapsed / 10))
        elapsed = status.elapsed;
    }
  }
  else
  {
    setCaption(STR_TITLE " " LANG_VERSION " " __VERSION);
    txTitle.setText(STR_TITLE " " LANG_VERSION " " __VERSION " - " __AUTHORS);
    txKbps.setText("");
    txKHz.setText("");

    QToolTip::add(&slVolume,   LANG_VOLUME);
    QToolTip::add(&slBalance,  LANG_CROSSFADE);
  }

  repaint(false);
}

void ControlPanel::slotError(const QString &error)
{
  txTitle.setText(error);
  txKbps.setText("");
  txKHz.setText("");
}

void ControlPanel::slotAbout(void)
{
  QMessageBox::about(this, STR_TITLE " " LANG_VERSION " " __VERSION,
                     "<big><b>" STR_TITLE " " LANG_VERSION " " __VERSION
#ifndef QT_NO_DEBUG
                     " [debug]"
#endif
                     "</b></big><br>"
                     __AUTHORS "<br>"
                     "This application uses the ZIPIO code Copyright (c) 1995, Edward B. Hamrick <br><br>"
                     "This program is distributed in the hope that it will be useful, "
                     "but <b>WITHOUT ANY WARRANTY</b>; without even the implied warranty of "
                     "<b>MERCHANTABILITY</b> or <b>FITNESS FOR A PARTICULAR PURPOSE</b>.  See the "
                     "GNU General Public License for more details.<BR><BR>"
                     "You should have received a copy of the GNU General Public License "
                     "along with this program; if not, write to the Free Software "
                     "Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA");
}

void ControlPanel::slotOpenSkin(void)
{
  QFile srcFile(QFileDialog::getOpenFileName());

  if (srcFile.exists())
  {
    char buffer[65536];
    QFile destFile(dir.absPath() + "/" + QFileInfo(srcFile.name()).fileName());

    if (srcFile.open(IO_ReadOnly))
    if (destFile.open(IO_WriteOnly))
    {

      for (unsigned i=0; i<srcFile.size(); i+=sizeof(buffer))
      {
        size_t len = srcFile.size() - i;
        if (len > sizeof(buffer))
          len = sizeof(buffer);

        srcFile.readBlock(buffer, len);
        destFile.writeBlock(buffer, len);
      }

      srcFile.close();
      destFile.close();

      Configuration::setSkin(QFileInfo(destFile.name()).fileName());
      this->skin = NULL; // slotRebuildSkinMenu will set this to the new skin
      slotRebuildSkinMenu();
      if (this->skin)
        this->skin->apply();
    }
  }

}

void ControlPanel::slotRebuildSkinMenu(void)
{
  QString skinName = Configuration::getSkin();
  QStringList list = dir.entryList("*.wsz", QDir::Files);

  if (list.count() != skinMenu.count() - 3)
  {
    // Clean up
    skinMenu.clear();
    if (true)
    for (std::vector<WASkin *>::iterator i=skins.begin(); i!=skins.end(); i++)
      delete *i;
    skins.clear();

    // Build
    skinMenu.insertItem(LANG_DEFAULT, &defaultSkin, SLOT(apply()));
    skinMenu.insertItem(LANG_OPEN, this, SLOT(slotOpenSkin()));
    skinMenu.insertSeparator();
    for (unsigned i=0; i<list.count(); i++)
    if (!list[i].startsWith("."))
    {
      WASkin *skin = new WASkin(list[i], dir.path() + "/" + list[i], this);
      skins.push_back(skin);
      skinMenu.insertItem(list[i].left(list[i].length()-4), skin, SLOT(apply()));

      // Check if this is the preferred skin
      if ((this->skin == NULL) && (list[i] == skinName))
        this->skin = skin;
    }
  }

  // This function may never return a NULL skin.
  if (skin == NULL)
    skin = &defaultSkin;
}

void ControlPanel::slotEjectPressed(bool)
{
  PlaylistManager::showme(&mpdCom, true);
}

void ControlPanel::slotLockVolume(void)
{
  volumeEdit = true;
}

void ControlPanel::slotReverseVolume(int vol)
{
  txTitle.setText(LANG_VOLUME ": " + QString::number(vol) + " %");
  mpdCom.volume(vol);
}

void ControlPanel::slotReleaseVolume(void)
{
  volumeEdit = false;
}

void ControlPanel::slotLockBalance(void)
{
  balanceEdit = true;
}

void ControlPanel::slotReverseBalance(int bal)
{
  txTitle.setText(LANG_CROSSFADE ": " + QString::number(bal * 0.1) + " sec");
  mpdCom.crossFade(bal);
}

void ControlPanel::slotReleaseBalance(void)
{
  balanceEdit = false;
}

void ControlPanel::slotLockSeek(void)
{
  seekEdit = true;
}

void ControlPanel::slotSetSeek(int)
{
}

void ControlPanel::slotReleaseSeek(void)
{
  mpdCom.seek(status.songIndex, slPosition.value() / 10);
  elapsed = slPosition.value();

  seekEdit = false;
}

void ControlPanel::playListClosed(void)
{
  btPl.setOn(false);
}

void ControlPanel::slotShowPlaylist(bool show)
{
  btPl.setOn(show);
  if (show)
    plw.show();
  else
    plw.hide();
}

void ControlPanel::slotShowSettings(void)
{
  if (configDialog.exec())
  {
    mpdCom.reconnectCom(Configuration::getServerIP(), Configuration::getServerPort());
    playlistMpdCom.reconnectCom(Configuration::getServerIP(), Configuration::getServerPort());
  }
}

void ControlPanel::Display::mousePressEvent(QMouseEvent *)
{
  reverseCounter = !reverseCounter;
}

void ControlPanel::closeEvent(QCloseEvent * event)
{
  Configuration::setGeometry("controlpanel", geometry());
  Configuration::setPlaylistVisible(btPl.isOn());
  Configuration::setCountMode(countMode);
  Configuration::setSkin(skin->getName());

  Configuration::close();
  QWidget::closeEvent(event);
}

void ControlPanel::paintEvent(QPaintEvent *)
{
  QPainter painter(this);

  switch (status.state)
  {
  case MpdCom::PLAY:  painter.drawPixmap(26, 28, skin->getState(0)); break;
  case MpdCom::STOP:  painter.drawPixmap(26, 28, skin->getState(2)); break;
  case MpdCom::PAUSE: painter.drawPixmap(26, 28, skin->getState(1));  break;
  }

  int elapsed = seekEdit ? slPosition.value() : this->elapsed;
  int tmp = (countMode ? status.songLength - elapsed : elapsed) / 10;
  int sec = tmp % 60;
  int min = tmp / 60;

  if (countMode) painter.drawPixmap(36, 26, skin->getNumber(11));
  painter.drawPixmap(48, 26, skin->getNumber(min / 10));
  painter.drawPixmap(60, 26, skin->getNumber(min % 10));
  painter.drawPixmap(78, 26, skin->getNumber(sec / 10));
  painter.drawPixmap(90, 26, skin->getNumber(sec % 10));
}

void ControlPanel::mouseMoveEvent(QMouseEvent *event)
{
  if (mouseDown)
  {
    int newx = actualX = actualX + event->globalX() - lastGlobalX;
    int newy = actualY = actualY + event->globalY() - lastGlobalY;

    // Snap to the edges of the screen
    #if QT_VERSION >= 300
      QDesktopWidget *widget = QApplication::desktop();
      QRect geo = widget->availableGeometry(this);
      if (abs(newx - geo.left()) < 12)                newx = geo.left();
      if (abs(newy - geo.top()) < 12)                 newy = geo.top();
      if (abs(newx + width() - geo.right()) < 12)     newx = geo.right() - width() + 1;
      if (abs(newy + height() - geo.bottom()) < 12)   newy = geo.bottom() - height() + 1;
    #endif

    // Move the widget
    move(newx, newy);
    lastGlobalX = event->globalX();
    lastGlobalY = event->globalY();
  }
}

void ControlPanel::mousePressEvent(QMouseEvent *event)
{
//  if (!plw.isHidden()) plw.raise();
  raise();

  if (event->button() == Qt::LeftButton)
  {
    lastX = event->x();
    lastY = event->y();
    lastGlobalX = event->globalX();
    lastGlobalY = event->globalY();

    if ((lastX >= 24) && (lastX <= 98) &&
        (lastY >= 25) && (lastY <= 39))
    {
      countMode = !countMode;
      repaint(true);
    }

    mouseDown = true;
  }
  else if (event->button() == Qt::RightButton)
    mainMenu.exec(event->globalPos());
}

void ControlPanel::mouseReleaseEvent(QMouseEvent *)
{
  mouseDown = false;
}

void ControlPanel::keyPressEvent(QKeyEvent * event)
{
  switch( event->key() )
  {
    case Qt::Key_Z      : mpdCom.previous(); break;
    case Qt::Key_X      : mpdCom.play(); break;
    case Qt::Key_C      : mpdCom.pause(); break;
    case Qt::Key_V      : mpdCom.stop(); break;
    case Qt::Key_B      : mpdCom.next(); break;
  }
}
