#ifndef __MPDCOM__H
#define __MPDCOM__H

#include <qobject.h>
#include <qsocket.h>
#include <qstringlist.h>
#include <queue>
#include <qtimer.h>

#include "mpdc.h"
#include "playlist.h"

/*! The communication with MPD is asynchronic which makes it possible to comunicate without
    stalling of the application. This class tries to wrap around differences in MPD versions
    as much as posible.
    \author     <a href="mailto:mpdc@hokitoki.dds.nl">ing. A.E. van Putten</a>
    \date       01-01-2004
    \brief      MpdCom class for communication with the Music Player Daemon.
 */
class MpdCom : public QObject
{
Q_OBJECT
public:
  enum                          responcetype { CONNECTING, SIMPLE, PLAYLIST, PLAYLIST_INFO, STATUS, FILELIST };
  enum                          mpdstate     { PLAY, STOP, PAUSE };

  /*! \author     <a href="mailto:mpdc@hokitoki.dds.nl">ing. A.E. van Putten</a>
      \date       01-01-2004
      \brief      Status class for storing status information.
   */
  class Status
  {
  public:
    inline                      Status(void) : volume(0), repeat(false), random(false), playlist(0),
                                               playlistLength(0), state(STOP), songIndex(0),
                                               elapsed(0), songLength(0), bitrate(0), songInfo(NULL) {}

    unsigned                    volume;
    bool                        repeat;
    bool                        random;
    unsigned                    playlist;
    int                         playlistLength;
    mpdstate                    state;
    unsigned                    songIndex;
    int                         elapsed;
    int                         songLength;
    int                         bitrate;
    Playlist::SongInfo        * songInfo;
    int                         xfade;
    unsigned                    samplerate, bits, channels;
  };

  /*! \author     <a href="mailto:mpdc@hokitoki.dds.nl">ing. A.E. van Putten</a>
      \date       01-01-2004
      \brief      Filelist class for storing directory content returned from the daemon.
   */
  class Filelist
  {
  public:
    /*! \author     <a href="mailto:mpdc@hokitoki.dds.nl">ing. A.E. van Putten</a>
        \date       01-01-2004
        \brief      Items for the playlist.
     */
    class ListItem
    {
    public:
                                ListItem( QString n, unsigned p ) : name(n), properties(p) {}

      QString                   name;                                           //!< The name of the item
      unsigned                  properties;                                     //!< Whether the item is a dir,file or playlist
    };

    enum                        { DIR, FILE, PLAYLIST };                        //!< For identification of the items
    typedef std::vector<ListItem *>::const_iterator iterator;                   //!< Iterator for traversing through the items
    inline iterator             begin() const                     { return list.begin(); } //!< Begin iterator
    inline iterator             end() const                       { return list.end(); }   //!< End iterator
    inline void                 clear()                           { list.clear(); }        //!< Clears the list

    inline Filelist::ListItem * operator [](unsigned i)           { return list[i]; }      //!< Operator for direct access to the items

    inline void                 add( QString n, unsigned p )      { list.push_back(new ListItem(n,p)); } //!< Add an item to the list

  private:
    std::vector<ListItem *>     list;                                           //!< The actual list
  };

protected:
  /*! \author     <a href="mailto:mpdc@hokitoki.dds.nl">ing. A.E. van Putten</a>
      \date       01-01-2004
      \brief      CommandItem class for storing command information.
   */
  class CommandItem
  {
  public:
    /*! Constructor, creates an instance of this class
        \param command          the command to send to MPD
        \param responce         the kind of responce to expect, defaults to SIMPLE
        \param send             wheather the command is send or not, defaults to false
    */
    inline CommandItem(QString command, responcetype responce = SIMPLE, bool send = false)
                     :command(command),
                      responce(responce),
                      send(send) {}

    QString                     command;                                        //!< The MPD command
    responcetype                responce;                                       //!< The expected responce
    bool                        send;                                           //!< Send or not
  };

public:
                                MpdCom( const char * hostname, int hostport, bool needsPlayList, QObject * ); //!< Constructs an instance of this class

  void                          reconnectCom(const char * hostname, int hostport); //!< Reconnects to a diffrent server.
  void                          disconnectCom(void);                            //!< Disconnect from the server.
  void                          connectCom(void);                               //!< Connect to the server.
  bool                          minVersion(QString);                            //!< Returns true if the connected version of MPD is higher of equel to the given version.

                                /*! Returns the version of MPD
                                */
  inline QString                getVersion(void)                                { return QString::number(majorVersion) + "." + QString::number(minorVersion) + "." + QString::number(buildVersion); }
                                /*! Starts a command list, all simple commands are stored and transfered when the endCommandList function is executed.
                                */
  inline void                   startCommandList(void)                          { buildingCommandList = true; commandList = "command_list_begin\n"; }
                                /*! End a command list and send the list to MPD
                                */
  inline void                   endCommandList(void)                            { buildingCommandList = false; commandList += "command_list_end"; postCommand(commandList); }

public slots:
  void                          previous(void);                                 //!< Plays the previous song
  void                          next(void);                                     //!< Plays the next song
  void                          stop(void);                                     //!< Stops playing
  void                          play(void);                                     //!< Starts playing
  void                          play(int);                                      //!< Starts playing a specific song
  void                          pause(void);                                    //!< Pauses the playback
  void                          seek(int songnr, int pos);                      //!< Seeks a position in a specific song
  void                          volume(int perc);                               //!< Sets the volume
  void                          repeat(bool);                                   //!< Toggles repeat
  void                          random(bool);                                   //!< Toggles random
  void                          update(void);                                   //!< Sends the update command
  void                          crossFade(int);                                 //!< Sets the crossfade amount

  void                          requestPlaylist(void);                          //!< Retrieves the playlist
  void                          loadPlaylist(QString, bool = false);            //!< Loads a playlist
  void                          clearPlaylist(void);                            //!< Clears the current playlist
  void                          deletePlaylist(QString);                        //!< Deletes a playlist
  void                          requestSongInfo(unsigned);                      //!< Requests songinformation about a specific song
  void                          requestStatusUpdate(void);                      //!< Requests the current status
  void                          requestDirectoryContent(QString dir = "");      //!< Requests the content of a directory
  void                          addFile(QString);                               //!< Adds a song to the playlist
  void                          removeFile(QString);                            //!< Removes a song from the playlist
  void                          swap(unsigned, unsigned);                       //!< Swaps two songs in the playlist

protected slots:
  void                          slotReadyRead(void);                            //!< Reads from the socket. This function handles all incoming data from MPD.
  void                          slotConnected(void);                            //!< Logs connection establishments
  void                          slotError(int);                                 //!< Logs connection errors
  void                          slotReconnect(void);                            //!< Reconnects with the server
  void                          slotConnectionClosed(void);                     //!< Logs connection closes

signals:
  void                          playlistUpdate(const Playlist &p);              //!< Sends a received playlist.       \param p the new playlist
  void                          statusUpdate(const MpdCom::Status &s);          //!< Sends new status information.    \param s the new status
  void                          error(const QString &e);                        //!< Sends error information.         \param e the error message
  void                          directoryContent(const MpdCom::Filelist &f);    //!< Sends received directorycontent. \param f the filelist

private:
  void                          setVersion(QString);
  void                          flushCommandQueue(void);

  std::queue<CommandItem *>     commandQueue;
  void                          postCommand(QString command, responcetype responce = SIMPLE );
  void                          sendData(void);

  QString                       hostname;
  int                           hostport;
  bool                          needsPlayList;
  QSocket                       socket;
  Playlist                      playlist;
  QTimer                        statusTimer, reconnectTimer;
  Playlist::SongInfo            songInfo;
  unsigned                      currentPlaylist;
  unsigned                      songIndex;
  Playlist::SongInfo          * si;
  Filelist                      filelist;
  QString                       data;
  unsigned                      majorVersion;
  unsigned                      minorVersion;
  unsigned                      buildVersion;
  int                           oldvol;
  unsigned                      playlistSize;
  bool                          buildingCommandList;
  QString                       commandList;
};

#endif
