#include <qapplication.h>

#include "playlistwindow.h"
#include "configuration.h"
#include "controlpanel.h"

/*! \param mpdcom               MPD communication
    \controlPanel               the controlpanel from which the playlist is called
*/
PlaylistWindow::PlaylistWindow(MpdCom * mpdcom, ControlPanel * controlPanel)
               :QWidget(NULL, NULL, Qt::WType_TopLevel),
                mainLayout(this, 3, 1),
                topLayout(this),
                bottomLayout(this),
                mpdcom(mpdcom),
                controlPanel(controlPanel),
                songList(mpdcom, this),
                editArtist(this),
                editTitle(this),
                labArtist(LANG_ARTIST ":", this),
                labTitle(LANG_TITLE ":", this),
                btAdd(LANG_ADD, this),
                btRemove(LANG_REMOVE, this),
                btClear(LANG_CLEAR, this),
                btRefresh(LANG_REFRESH, this),
//                playlistManager(NULL),
                currentPlaying(NULL)
{
  setGeometry(Configuration::getGeometry("playlistwindow"));

  topLayout.addWidget(&labArtist);
  topLayout.addWidget(&editArtist);
  topLayout.addWidget(&labTitle);
  topLayout.addWidget(&editTitle);
  topLayout.setStretchFactor(&editArtist, 1);
  topLayout.setStretchFactor(&editTitle, 1);
  dbgConnect( &editArtist, SIGNAL(textChanged(const QString &)), this, SLOT(slotFind(void)) );
  dbgConnect( &editTitle, SIGNAL(textChanged(const QString &)), this, SLOT(slotFind(void)) );
  mainLayout.addLayout(&topLayout, 1, 1);

  songList.addColumn(LANG_INDEX);
  songList.addColumn(LANG_ARTIST);
  songList.addColumn(LANG_TITLE);
  songList.setColumnWidthMode(1, QListView::Manual);
  songList.setColumnWidth(1, 200);
  songList.setColumnWidth(2, 285);
  songList.addColumn(LANG_LENGTH);
  songList.setAllColumnsShowFocus(true);
  #ifdef QT_NO_DEBUG
    songList.setColumnWidthMode(0, QListView::Manual);
    songList.setColumnWidth(0, 0);
    songList.setSorting(1);
  #endif
  mainLayout.addWidget(&songList, 2, 1);

  bottomLayout.addWidget(&btRemove);
  dbgConnect( &btRemove, SIGNAL(pressed()), this, SLOT(slotRemoveSong()));
  bottomLayout.addWidget(&btAdd);
  dbgConnect( &btAdd, SIGNAL(pressed()), this, SLOT(slotAddSong()) );
  bottomLayout.addWidget(&btClear);
  dbgConnect( &btClear, SIGNAL(pressed()), this, SLOT(slotClearPlaylist()) );

  bottomLayout.addStretch();
  bottomLayout.addWidget(&btRefresh);
  dbgConnect( &btRefresh, SIGNAL(pressed()), this, SLOT(slotRefreshPlaylist()) );
  mainLayout.addLayout(&bottomLayout, 3, 1);

  dbgConnect( mpdcom, SIGNAL(playlistUpdate(const Playlist &)), this, SLOT(slotPlaylistUpdate(const Playlist &)) );
  dbgConnect( &songList, SIGNAL(doubleClicked(QListViewItem *)), this, SLOT(slotDoubleClicked(QListViewItem *)) );
  dbgConnect( &songList, SIGNAL(returnPressed(QListViewItem *)), this, SLOT(slotDoubleClicked(QListViewItem *)) );
  dbgConnect( mpdcom, SIGNAL(statusUpdate(const MpdCom::Status &)), this, SLOT(slotReceiveUpdate(const MpdCom::Status &)) );
}

void PlaylistWindow::show(void)
{
  mpdcom->connectCom();
  QWidget::show();
}

void PlaylistWindow::hide(void)
{
  mpdcom->disconnectCom();
  QWidget::hide();
}

PlaylistWindow::~PlaylistWindow(void)
{
}

void PlaylistWindow::slotPlaylistUpdate( const Playlist & playlist )
{
  if(songList.canUpdate())
  {
    debug( "Update playlist" );
    currentPlaying = NULL;
    editArtist.setText("");
    editTitle.setText("");
    songList.clear();
    for(Playlist::iterator i=playlist.begin(); i!=playlist.end(); i++)
    {
      new PlaylistViewItem( &songList,
                          ("0000" + QString::number((*i)->songIndex)).right(5),
                          (*i)->artist,
                          (*i)->title,
                          secToHMS((*i)->songLength) );
    }
  }
}

void PlaylistWindow::slotDoubleClicked( QListViewItem * item )
{
  mpdcom->play(item->text(0).toUInt());
}

void PlaylistWindow::slotRefreshPlaylist( void )
{
  mpdcom->requestPlaylist();
}

void PlaylistWindow::slotAddSong( void )
{
//  if( !playlistManager ) playlistManager = new PlaylistManager(mpdcom);
  PlaylistManager::showme(mpdcom);
  btAdd.setDown(false);
}

void PlaylistWindow::slotRemoveSong(void)
{
  QApplication::sendEvent(&songList, new QKeyEvent(QEvent::KeyPress, Qt::Key_Delete, 127, 0));
}

void PlaylistWindow::slotClearPlaylist(void)
{
  mpdcom->clearPlaylist();
}

void PlaylistWindow::slotFind( void )
{
  #if QT_VERSION >= 300
    if( (editArtist.text().length() > 2) || (editArtist.text().length() == 0)
    ||  (editTitle. text().length() > 2) || (editTitle .text().length() == 0) )
    {
      QListViewItem * i = songList.firstChild();
      if(i)
      do
      {
        i->setVisible( (i->text(1).contains(editArtist.text(),false)) &&
                       (i->text(2).contains(editTitle.text(),false)) );
      }while((i = i->nextSibling()) > 0);
    }
  #endif
//    songList.clear();
//    for(Playlist::iterator i=playlist.begin(); i!=playlist.end(); i++)
//    {
//      if( ((*i)->artist.contains(editArtist.text(), false)) && ((*i)->title.contains(editTitle.text(), false)) )
//        new PlaylistViewItem( &songList, ("0000" + QString::number((*i)->songIndex)).right(5), (*i)->artist, (*i)->title, secToHMS((*i)->songLength) );
//    }
//
//    markPlaying((PlaylistViewItem *)songList.findItem( ("0000" + QString::number(status.songIndex)).right(5), 0 ), false);
//  }
}

void PlaylistWindow::markPlaying(PlaylistViewItem * item, bool makevisable = true)
{
  if(!item) return;
  if(currentPlaying){
    currentPlaying->setPlaying(false);
    songList.repaintItem(currentPlaying);
  }
  currentPlaying = item;
  currentPlaying->setPlaying(true);
  if(makevisable) songList.ensureItemVisible(item);
  songList.repaintItem(currentPlaying);
}

void PlaylistWindow::slotReceiveUpdate(const MpdCom::Status & stat)
{
  if( stat.state == MpdCom::PLAY )
  if( songList.childCount() > 0 )
  if( stat.songIndex != status.songIndex )
  {
    #if QT_VERSION >= 300
      markPlaying((PlaylistViewItem *)songList.findItem( ("0000" + QString::number(stat.songIndex)).right(5), 0 ));
    #endif
    status = stat;
  }
}

void PlaylistWindow::PlaylistViewItem::paintCell ( QPainter * p, const QColorGroup & cg, int column, int width, int align )
{
  if(!isPlaying() || isSelected())
  {
    QListViewItem::paintCell(p,cg,column,width,align);
    return;
  }

  int half = height()/2;
  int color;

  OLD_FOR_SCOPE_PATCH
  for(int i=0; i<half; i++)
  {
    color = 150+(i*(155/height()));
    p->setPen(QColor(color,color,color));
    p->drawLine(0,i,width,i);
  }

  OLD_FOR_SCOPE_PATCH
  for(int i=height(); i>=half; i--)
  {
    color = 205-( (i-half) * (155/height()) );
    p->setPen(QColor(color,color,color));
    p->drawLine(0,i,width,i);
  }

  p->setPen(QColor(0,0,0));
  p->drawText(1,height()-4, text(column));
}

void PlaylistWindow::PlayListView::keyPressEvent(QKeyEvent * event)
{
  QListViewItem * item = selectedItem();
  if(item)
  switch( event->key() )
  {
    case Qt::Key_Delete : mpdCom->removeFile(item->text(0));
                          break;
    case Qt::Key_Z      : mpdCom->previous(); break;
    case Qt::Key_X      : mpdCom->play(); break;
    case Qt::Key_C      : mpdCom->pause(); break;
    case Qt::Key_V      : mpdCom->stop(); break;
    case Qt::Key_B      : mpdCom->next(); break;
    default:              QListView::keyPressEvent(event);
  }
}

void PlaylistWindow::PlayListView::contentsMousePressEvent(QMouseEvent * event)
{
  debug( "PlaylistWindow::PlayListView::mousePressEvent" );
  QListView::contentsMousePressEvent(event);

  dragging = (PlaylistViewItem *)itemAt(contentsToViewport(event->pos()));
  if(dragging)
  {
    debug( dragging->text(0).latin1() );
    prevMouseY = event->globalY();
  }
}

void PlaylistWindow::PlayListView::contentsMouseMoveEvent ( QMouseEvent * event)
{
  QListView::contentsMouseMoveEvent(event);

  if(event->state() == Qt::LeftButton)
  {
    PlaylistViewItem * current = (PlaylistViewItem *)itemAt(contentsToViewport(event->pos()));
    if(current && dragging)
    if(current != dragging)
    {
      noUpdate = true;
      if( event->globalY() < prevMouseY )
      {
        QListViewItem * prev = current->itemAbove();
        if(prev)
        {
          dragging->moveItem(prev);
          mpdCom->swap(dragging->text(0).toUInt(), prev->text(0).toUInt());
        }
        debug( "Moving item up " << dragging->text(0).toUInt() << "/" << prev->text(0).toUInt() );
      }else{
        dragging->moveItem(current);
        mpdCom->swap(dragging->text(0).toUInt(), current->text(0).toUInt());
        debug( "Moving item down " << dragging->text(0).toUInt() << "/" << current->text(0).toUInt() );
      }
      prevMouseY = event->globalY();

//      takeItem(dragging);
//      current->insertItem(dragging);
//      dragging = current;
    }
  }
}

void PlaylistWindow::PlayListView::contentsMouseReleaseEvent(QMouseEvent * event)
{
  debug( "PlaylistWindow::PlayListView::mouseReleaseEvent" );
  QListView::contentsMouseReleaseEvent(event);

  noUpdate = false;
  dragging = NULL;
}

void PlaylistWindow::closeEvent(QCloseEvent * event)
{
  Configuration::setGeometry("playlistwindow", geometry());
  controlPanel->playListClosed();

  QWidget::closeEvent(event);
}
